"""ZWiki product

This product allows you to build wiki webs with zope objects.
It includes a sample wikiweb to help you get started quickly.
There is a demo & documentation site running at:

 http://joyful.com/zwiki/

You are responsible for security; zwiki may contain unforeseen risks.
I appreciate all feedback, bugreports, fixes, responses to the ??
comments in the code, re-implementations, etc.

(c) 1999,2000 Simon Michael <simon@joyful.com> & contributors
This product is available under the GPL.
All rights reserved, all disclaimers apply, etc.

?? marks questions and dubious areas

"""

__version__="$Revision: 0.12 $"[11:-2]

from Globals          import HTMLFile, MessageDialog
from DocumentTemplate import HTML
from OFS.DTMLDocument import DTMLDocument
from OFS.Document     import Document
from App.Common       import absattr
from urllib           import quote, unquote
from DocumentTemplate.DT_Util import html_quote
import re
from StructuredText   import html_with_references
from wwml             import translate_WMML
from string           import split,join,strip
# --- geoffg's managed_mode additions
from DateTime         import DateTime
# end geoffg's managed_mode additions
#disable Catalog support for the moment
#from Products.ZCatalog.CatalogAwareness import CatalogAware
from types import *
import Acquisition


intl_char_entities = (
#    ('\253', '&#171;'),                 ##
#    ('\273', '&#187;'),                 ##
    ('\300', '&Agrave;'),               ##
    ('\302', '&Acirc;'),                ##
    ('\311', '&Eacute;'),               ##
    ('\312', '&Ecirc;'),                ##
    ('\316', '&Icirc;'),                ##
    ('\324', '&Ocirc;'),                ##
    ('\333', '&Ucirc;'),                ##
    ('\340', '&agrave;'),               ##
    ('\342', '&acirc;'),                ##
    ('\347', '&ccedil;'),               ##
    ('\350', '&egrave;'),               ##
    ('\351', '&eacute;'),               ##
    ('\352', '&ecirc;'),                ##
    ('\356', '&icirc;'),                ##
    ('\364', '&ocirc;'),                ##
    ('\371', '&ugrave;'),               ##
    ('\373', '&ucirc;'),                ##
#    ('(r)', '&#174;'),
#    ('(tm)', '&#153;'),
#    ('^s', '&nbsp;')
)

# regular expressions
# worth compiling these ??

# regexps for international-character links
# this doesn't help because we store them as entities
# (to protect from loss due to.. browser textareas I think)
# supporting this may be tricky
#IU = intlupper        = '\300\302\311\312\316\324\333'
#IL = intllower        = '\340\342\347\350\351\352\356\364\371\373'
#IA = intlall          = IU+IL
#urllink          = r'["=]?((http|ftp|mailto|file):[A-Za-z0-9/:@_%~\.\-\?'+IA+']+)'
#wikiname1        = r'[A-Z'+IU+']+[a-z'+IL+']+[A-Z'+IU+'][A-Za-z'+IA+']*'
#wikiname2        = r'[A-Z]'+IU+'[A-Z'+IU+']+[a-z'+IL+'][A-Za-z'+IA+']*'

strip_cr         = r'\r(\n)'
bracketedexpr    = r'\[([^\]0-9][^]]*)\]' # trying to co-exist with stx's references
underlinedexpr   = r'_([^_]+)_'
urllink          = r'["=]?((http|ftp|mailto|file):[A-Za-z0-9/:@_%~#=\.\-\?]+)'
# allow digits at the end of wikinames ?
#wikiname1        = r'[A-Z]+[a-z]+[A-Z][A-Za-z]*'
#wikiname2        = r'[A-Z][A-Z]+[a-z][A-Za-z]*'
wikiname1        = r'\b[A-Z]+[a-z]+[A-Z][A-Za-z]*[0-9]*'
wikiname2        = r'\b[A-Z][A-Z]+[a-z][A-Za-z]*[0-9]*'
simplewikilink   = r'!?(%s|%s)' % (wikiname1, wikiname2)
#wikilink         = r'!?(%s|%s|%s|%s|%s)' % (wikiname1,
wikilink         = r'!?(%s|%s|%s|%s)' % (wikiname1,
                                            wikiname2,
                                            bracketedexpr,
#darn!                                           underlinedexpr,
                                            urllink)
interwikilink    = r'!?((?P<local>%s):(?P<remote>[\w]+))' % (wikilink)
remotewikiurl    = r'(?m)RemoteWikiURL[:\s]*(.*)$'
protected_line   = r'(?m)^!(.*)$'


DEFAULT_PAGE_TYPE = 'structuredtext'


class ZWikiPage(DTMLDocument): #, CatalogAware):
    """A ZWikiPage is a DTML Document which knows how to render itself in 
       various wiki styles, and provides some utility methods to support
       wiki-building.
    """
    meta_type="ZWiki Page"
    icon     ="misc_/ZWiki/ZWikiPage_icon"
    page_type=DEFAULT_PAGE_TYPE

    # how would you append to the superclass' _properties here ??  #'
    _properties=({'id':'title', 'type': 'string', 'mode':'w'},
                 {'id':'page_type', 'type': 'string', 'mode': 'w'},
# ken's parenting code:
                 {'id':'parents', 'type': 'lines', 'mode': 'w'},
# ^^^ ken's parenting code
                 # --- geoffg's managed_mode additions
                 {'id':'managed_mode', 'type': 'boolean', 'mode': 'w'},
                 {'id':'sticky_owner', 'type': 'boolean', 'mode': 'w'},
                 {'id':'page_owner', 'type': 'string', 'mode': 'w'},
                 {'id':'page_status', 'type': 'string', 'mode': 'w'}
                 # end geoffg's managed_mode additions
                 )

    __ac_permissions__=(
        # 'manage_edit','manage_upload' already part of Change DTML Doc permission - ok here ??
        ('Change ZWiki Pages', ('edit','PUT','manage_edit','manage_upload')), 
        # --- geoffg's managed_mode additions (approve and manage pages)
        ('Approve ZWiki Pages', ()),
        ('Manage ZWiki Pages', ()),
        # end geoffg's managed_mode additions
        # add a View permission here ??
        # useful to define default roles ??
        )

# ken's parenting code:
    parents = None
# ^^^ ken's parenting code
    # --- geoffg's managed_mode additions
    managed_mode = None
    sticky_owner = None
    page_owner = ''
    page_status = ''
    # end geoffg's managed_mode additions
                    
    def __call__(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        """Render a zwiki page, with standard header & footer
        """
        # dreadful-hacks-dept.
        # define this here so header & footer methods can get at
        # the page attributes. why needed ?? better solution ??
        kw['this'] =self

        # backwards compatibility
        # XXX Note that the objects don't get very far unpickling, some
        #     by-hand adjustment via command-line interaction is necessary to
        #     get them over the transition, sigh. (?? check what this means)
# ken's parenting code:
        if (not hasattr(self, 'parents')) or self.parents is None:
            self.parents = []
# ^^^ ken's parenting code
        if not hasattr(self, 'page_type'):
            self.page_type = DEFAULT_PAGE_TYPE
        elif self.page_type == 'Structured Text':
            self.page_type = 'structuredtext'
        elif self.page_type == 'structuredtext_dtml':
            self.page_type = 'structuredtextdtml'
        elif self.page_type[0:4] == 'HTML':
            self.page_type = 'html'
        elif  self.page_type == 'html_dtml':
            self.page_type = 'htmldtml'
        elif self.page_type == 'Classic Wiki':
            self.page_type = 'classicwiki'
        elif self.page_type == 'Plain Text':
            self.page_type = 'plaintext'

        # prepend wiki header
        if REQUEST is not None:
            if hasattr(self, 'standard_wiki_header'):
                hdr = apply(self.standard_wiki_header.__call__,
                            (client, REQUEST, RESPONSE), kw)
            else:
                hdr = default_wiki_header(self,REQUEST) # ?? guessing
        else: hdr = ''

        # call render_<page_type> to render the body
        r = 'render_' + self.page_type
        if hasattr(self, r):
            r = getattr(self,r)
        else:
            r = self.render_plaintext

        body = apply(r,(client, REQUEST, RESPONSE), kw)

        # append wiki footer
        if REQUEST is not None:
            if hasattr(self, 'standard_wiki_footer'):
                ftr = apply(self.standard_wiki_footer.__call__,
                            (client, REQUEST, RESPONSE), kw)
            else:
                ftr = default_wiki_footer(self,REQUEST) # ?? guessing
        else: ftr = ''

	if RESPONSE is not None:
	    RESPONSE.setHeader('Content-Type', 'text/html')
        return hdr + body + ftr


    # built-in render methods - add new ones with python/external/dtml
    # methods. Their names correspond with the page_type attribute. NB
    # if the name contains "dtml", we will do DTML validation after an
    # edit
    #
    # I inlined the code from _wikiize() for maximum flexibility

    def render_structuredtextdtml(self, client=None, REQUEST={},
                                  RESPONSE=None, **kw): 
        # DTML + structured text + wiki links + HTML
        t = apply(DTMLDocument.__call__,(self, client, REQUEST, RESPONSE), kw)
        t = str(html_with_references(t, level=3))
        t = re.sub(protected_line, self._protect_line, t)
        t = re.sub(interwikilink, self._interwikilink_replace, t)
        t = re.sub(wikilink, self._wikilink_replace, t)
        return t

    def render_structuredtext(self, client=None, REQUEST={},
                              RESPONSE=None, **kw):
        # structured text + wiki links + HTML
        t = self.raw
        t = str(html_with_references(t, level=3))
        t = re.sub(protected_line, self._protect_line, t)
        t = re.sub(interwikilink, self._interwikilink_replace, t)
        t = re.sub(wikilink, self._wikilink_replace, t)
        return t

    def render_htmldtml(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        # DTML + wiki links + HTML
        t = apply(DTMLDocument.__call__,(self, client, REQUEST, RESPONSE), kw)
        t = re.sub(protected_line, self._protect_line, t)
        t = re.sub(interwikilink, self._interwikilink_replace, t)
        t = re.sub(wikilink, self._wikilink_replace, t)
        return t

    def render_html(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        # wiki links + HTML
        t = self.raw
        t = re.sub(protected_line, self._protect_line, t)
        t = re.sub(interwikilink, self._interwikilink_replace, t)
        t = re.sub(wikilink, self._wikilink_replace, t)
        return t

    def render_classicwiki(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        # classic wiki formatting + (inter)wiki links
        t = html_quote(self.raw)
        t = translate_WMML(t)
        t = re.sub(interwikilink, self._interwikilink_replace, t)
        t = re.sub(simplewikilink, self._simplewikilink_replace, t)
        return t

    def render_plaintext(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        # fixed-width plain text with no surprises
        t = "<pre>\n" + html_quote(self.raw) + "\n</pre>\n"
        return t

    # ?? tasty spaghetti regexps! better suggestions welcome

    def _protect_line(self, match):
        """protect an entire line from _wikilink_replace, by
           protecting all its wikilink occurrences
        """
        return re.sub(wikilink, r'!\1', match.group(1))

    def _simplewikilink_replace(self, match):
        """replace an occurrence of simplewikilink with a suitable hyperlink
        """
        # matches beginning with ! should be left alone
        if re.match('^!',match.group(0)):
            return match.group(1)

        m = match.group(1)

        # if something of this name exists, link to it;
        # otherwise, provide a "?" creation link
        if hasattr(self.aq_parent, m): 
            return '<a href="%s">%s</a>' % (quote(m), m)
        else:
            return '%s<a href="%s/new?id=%s">?</a>' % (m, quote(self.id()),
                                                       quote(m))
        
    def _wikilink_replace(self, match):
        """replace an occurrence of wikilink with a suitable hyperlink
        """
        # matches beginning with ! should be left alone
        if re.match('^!',match.group(0)):
            return match.group(1)

        m = match.group(1)

        # strip any enclosing [] or __
        # display these as a visual cue ??
        if re.match(bracketedexpr,m):
            m = re.sub(bracketedexpr, r'\1', m)
        elif re.match(underlinedexpr,m):
            m = re.sub(underlinedexpr, r'\1', m)
        
        if re.match(urllink,m):
            if re.match('^["=]',m):                                   # "
                # a url preceded by " or = should probably be left alone
                return m
            else:
                return '<a href="%s">%s</a>' % (m, m)
        elif hasattr(self.aq_parent, m):
            # if something of this name exists, link to it;
            return '<a href="%s">%s</a>' % (quote(m), m)
        else:
            # otherwise, provide a "?" creation link
            return '%s<a href="%s/new?id=%s">?</a>' % (m, quote(self.id()), quote(m))
        
    def _interwikilink_replace(self, match):
        """replace an occurrence of interwikilink with a suitable hyperlink
        """
        # matches beginning with ! should be left alone
        if re.match('^!',match.group(0)): return match.group(1)

        localname  = match.group('local')
        remotename = match.group('remote') # named groups come in handy here!

        # look for a RemoteWikiURL definition
        if hasattr(self.aq_parent, localname): 
            localpage = getattr(self.aq_parent,localname)
            m = re.search(remotewikiurl, str(localpage))
            if m is not None:
                remoteurl = m.group(1)
                
                # we have a valid inter-wiki link
                link = '<a href="%s%s">%s:%s</a>' % \
                       (remoteurl, remotename, localname, remotename)
                # protect it from any further wiki-ization
                return re.sub(wikilink, r'!\1', link)

        # otherwise, leave alone
        return match.group(0)

    
    def new(self, id, title='', REQUEST=None, submit=None):
        """Create a new ZWikiPage and display it.
        """
        self._validateProxy(REQUEST)    # correct ??

        # set up default content
        if hasattr(self.aq_parent, 'standard_wiki_page'):
            t = apply(self.aq_parent.standard_wiki_page.__call__,
                      (None, REQUEST, REQUEST.RESPONSE))
        else:
            t = default_wiki_page

        # create page
        ob=ZWikiPage(t, __name__=id)
        ob.title=title
        id = self.aq_parent._setObject(id,ob)
# ken's parenting code:
        if ob.parents is None:
            ob.parents = [self.id()]
        else:
            ob.parents.append(self.id())
# ^^^ ken's parenting code

        # --- geoffg's managed_mode additions
        # New page sets managed_mode and sticky_owner from its parent i.e. self
        ob.managed_mode = self.managed_mode
        ob.sticky_owner = self.sticky_owner
        # end geoffg's managed_mode additions

        # add it to the catalog
        # can't use ob.index_object(), because ob doesn't have the
        # zpublisher/acquisition wrapper I think
#        if hasattr(self, self.default_catalog):
#            url = join(split(self.url(),'/')[:-1],'/') + '/' + id   # !!??
#            getattr(self, self.default_catalog).catalog_object(ob, url)

        if REQUEST is not None:
            # redirect browser to the new page
            # all these redirects.. better/safer way ??
            u=REQUEST['URL2']
            REQUEST.RESPONSE.redirect(u + '/' + quote(id))

    # jim's http & ftp edit conflict safety belts
    def manage_FTPget(self):
        "Get source for FTP download"
        return "Wiki-Safetybelt: %s\n\n%s" % (
            self.editTimestamp(), self.read())

    def PUT(self, REQUEST, RESPONSE):
        """Handle HTTP PUT requests."""
        self.dav__init(REQUEST, RESPONSE)
        body=REQUEST.get('BODY', '')
        self._validateProxy(REQUEST)

        m=re.match(r'Wiki-Safetybelt: ([0-9]+[.][0-9]+)[\n][\n]?', body)
        if m:
            self.checkEditTimeStamp(m.group(1))
            body=body[m.span()[1]:]            

        self.munge(body)
        RESPONSE.setStatus(204)
        return RESPONSE

    def checkEditTimeStamp(self, timeStamp):
        if timeStamp is not None and timeStamp != self.editTimestamp():
            raise 'EditingConflict', (
                '''Someone has edited this page since you loaded the
                page for editing.<p>
                Try editing the page again.
                ''')
                                                            

    def editTimestamp(self): return str(self._p_mtime)
    def edit(self, text=None, REQUEST=None, submit=None, type=None,
             timeStamp=None):
        """change a ZWikiPage's content and/or markup type
        """
        self._validateProxy(REQUEST)    # correct ??

        self.checkEditTimeStamp(timeStamp)

        if type is not None:
            self.page_type = type

        #######
        # --- geoffg's managed_mode additions
        ###
        # edit form has hidden fields:
        #   'check_managed_mode' - look for existence of 'new_managed_mode'
        #   'check_sticky_owner' - look for existence of 'new_sticky_owner'
        #   'check_page_owner' - look for existence of:
        #     'clear_page_owner' or 'retain_page_owner' or 'become_page_owner'
        # also might have radiobuttons
        #   'page_approval' values ['none', 'approve', 'reject']
        #   'request_review' values ['none', 'edit', 'review']
        ###
        # this page object has fields
        #   'managed_mode' values ['1', None]
        #   'sticky_owner' values ['1', None]
        #   'page_owner' string value
        #   'page_status' values ['', 'draft', 'review', 'approved', 'rejected']
        if REQUEST is not None:
          if REQUEST.has_key('check_managed_mode'):
            if REQUEST.has_key('new_managed_mode'):
              self.managed_mode='1'
            else:
              self.managed_mode=None
          if self.managed_mode:
            if REQUEST.has_key('check_sticky_owner'):
              if REQUEST.has_key('new_sticky_owner'):
                self.sticky_owner='1'
              else:
                self.sticky_owner=None
            s = ''
            s1 = str(REQUEST['AUTHENTICATED_USER']) + ' on ' + DateTime.ISO(DateTime())
            if REQUEST.has_key('new_text'):
              if REQUEST['new_text'] <> '': 
                text = text + REQUEST['new_text']
                s = 'Contributed by ' + s1
            else:
              s = 'Edited by ' + s1
            bgc = '#f0f0f0'     # light silver
            ###
            # Approval takes precedence over review requests
            # If page_approval radiobuttons were active, then ...
            #   you can approve a page that is in review status
            #   and reject a page that is in review or approved status
            if REQUEST.has_key('page_approval'):
              if REQUEST['page_approval'] == 'approve':
                if self.page_status == 'review':
                  self.page_status = 'approved'
                s = 'Approved by ' + s1
                bgc = '#99ff66' # light green
              elif REQUEST['page_approval'] == 'reject':
                if self.page_status in ['review', 'approved']:
                  self.page_status = 'rejected'
                s = 'Rejected by ' + s1
                bgc = '#ffcc33' # light red
            ###
            # Look at review requests (in practice only if no approval request in previous para)
            # If request_review radiobuttons were active ...
            #   you can ask for review or edit from any status
            if REQUEST.has_key('request_review'):
              if REQUEST['request_review'] == 'review':
                self.page_status = 'review'
              if REQUEST['request_review'] == 'edit':
                self.page_status = 'draft'
            ###
            # If edit marker hasn't already been filled in by approval activity
            # then append one
            if s <> '':  
              text = text + '\n<p><table class=mm width="100%" bgcolor=' 
              text = text + bgc 
              text = text + ' cellspacing="0" frame="box" border="1"><tr><td>'
              text = text + s + '</td></tr></table></p>\n'
            if REQUEST.has_key('check_page_owner'):
              if REQUEST['check_page_owner']=='clear_page_owner':
                self.page_owner=''
              elif REQUEST['check_page_owner']=='become_page_owner':
                self.page_owner=str(REQUEST['AUTHENTICATED_USER'])
              # else do nothing to page owner
          else:
            self.page_owner=''
            self.page_status=''
            self.sticky_owner=None
        ###
        # end geoffg's managed_mode additions
        #######

        if text is not None:
            text = self._edit_cleanups(text)
            if re.search(r'(?i)dtml',self.page_type):
                # for DTML page types, test for DTML problems - zope
                # will roll back this transaction, and the user will
                # get an appropriate error
                self.munge(text)
                apply(DTMLDocument.__call__,(self, self.aq_parent,
                                             REQUEST, REQUEST.RESPONSE))
            else:
                self.raw = text


        # update this page's catalog entry
#        self.reindex_object()

        if REQUEST is not None:
            # force a browser page refresh
            try: u=self.DestinationURL()
            except: u=REQUEST['URL1']
            REQUEST.RESPONSE.redirect(u)

    def _edit_cleanups(self, t=None):
        """do some cleanups on new page text
        """
        # strip any browser-appended ^M's
        t = re.sub(strip_cr, r'\1', t)

        # convert international characters to HTML entities for safekeeping
        for c,e in intl_char_entities:
            t = re.sub(c, e, t)

        # ?? here's the place to strip out any disallowed html/scripting elements 

        return t
        
    src__roles__ = None
    def src(self, REQUEST=None, RESPONSE=None):
        """Permission-free version of document_src""" # why ??
        if RESPONSE is not None:
            RESPONSE.setHeader('Content-Type', 'text/plain')
        return self.read()

    # url utilities to make links easier
    # easier way to get this information ??
    def wiki_page_url(self):
        """return the url path for the current wiki page"""

        o = self
        url = []
        while hasattr(o,'id'):
            url.insert(0,absattr(o.id))
            o = getattr(o,'aq_parent', None)
        return quote('/' + join(url[1:],'/'))

    def wiki_base_url(self):
        """return the base url path for the current wiki"""

        u = self.wiki_page_url()
        u = split(u,'/')
        u = u[:-1]
        u = join(u,'/')
        return quote(u)

    # another fine safety-belt from jim:
    # only access unprotected resources when using DTML
    # I'm hazy on this.. let's see what it does :)
    def validate(self, inst, parent, name, value, md):
 
         try:
             if (name[:3]=='aq_' and
                 name != 'aq_parent' and name != 'aq_explicit'):
                 return 0
         except: pass # name might not be a string!
 
         # Try to get roles
         if hasattr(value, '__roles__'): roles=value.__roles__
         else:
             if hasattr(parent,'__roles__'): roles=parent.__roles__
             elif hasattr(parent, 'aq_acquire'):
                 try: roles=parent.aq_acquire('__roles__')
                 except AttributeError:
                     if hasattr(inst, 'aq_base'): inst=inst.aq_base
                     if hasattr(parent, 'aq_base'): parent=parent.aq_base
                     return inst is parent
             else:
                 if hasattr(inst, 'aq_base'): inst=inst.aq_base
                 if hasattr(parent, 'aq_base'): parent=parent.aq_base
                 return inst is parent
             value=parent
             
         if roles is None or ("Anonymous" in roles): return 1
 
         if inst is parent:
             raise 'Unauthorized', (
                 'You are not authorized to access <em>%s</em>.' % name)
 
         return 0
 
# ken's parenting code:
    def reparent(self, parents=None, REQUEST=None):
        """Reset parents property according to request."""
        if parents is None:
            parents = REQUEST.get('parents', None)
        if parents is None:
            self.parents = []
        else:
            if type(parents) != ListType:
                parents = [parents]
            self.parents = parents
        if REQUEST is not None:
            REQUEST.RESPONSE.redirect(REQUEST['URL1'] + "/backlinks")


    def get_ancestors(self):
        """Return a trimmed nesting structure indicating this page's ancestors.

        See the WikiNesting docstring for nestings structure description."""
        container = self.aq_parent
        ancestors = {}
        offspring = {}
        tops = {}                       # Ancestors that have no parents.
        todo = {self.id(): None}
        while todo:
            doing = todo
            todo = {}
            for i in doing.keys():
                if ancestors.has_key(i):
                    continue            # We've already collected this one.
                else:
                    if not hasattr(container, i):
                        continue        # Absent - ignore it.
                    ancestors[i] = None
                    obj = container[i]
                    if not hasattr(obj, 'parents'):
                        continue
                    parents = obj.parents
                    if type(parents) != ListType: # SKWM I think ken's zope remembers [] better
                        parents = []
                    if parents:
                        for p in parents:
                            if offspring.has_key(p):
                                offspring[p].append(i)
                            else:
                                offspring[p] = [i]
                            todo[p] = None
                    else: tops[i] = None

        # Ok, now go back down, unravelling each forebear only once:
        tops = tops.keys()
        tops.sort
        did = {}; got = []
        for t in tops:
            got.append(descend_ancestors(t, ancestors, did, offspring))
        return got

    def context(self, REQUEST=None, with_siblings=0):
        """Return HTML showing this page's parents and siblings."""
        myid = self.id()
        if with_siblings:
            nesting = WikiNesting(self.aq_parent).get_up_and_back(myid)
        else:
            nesting = self.get_ancestors()
            if (len(nesting) == 0  or
                (len(nesting) == 1 and len(nesting[0]) == 1)):
                return "&nbsp;"
#        return present_nesting(myid, nesting, self.aq_parent.absolute_url())
        return present_nesting(self, nesting, self.wiki_base_url())     # SKWM allow path trickery
                
    def offspring(self, REQUEST=None):
        """Return a presentation of all my offspring."""
        myid = self.id()
        nesting = WikiNesting(self.aq_parent)
        return present_nesting(self, nesting.get_offspring([myid]),
#                               self.aq_parent.absolute_url())
                               self.wiki_base_url())     # SKWM allow path trickery
    def map(self, REQUEST=None):
        """Present the nesting layout of the entire wiki, showing:

        - All the independent nodes, ie those without parents or children, and

        - All the branches in the wiki - from the possibly multiple roots."""
        import string
        map = WikiNesting(self.aq_parent).get_map()
        singletons = []
        combos = []
#        rel = self.aq_parent.absolute_url()
        rel = self.wiki_base_url()     # SKWM allow path trickery
        for i in map:
            if type(i) == StringType:
                singletons.append('<a href="%s/%s">%s</a>'
                                  % (rel, quote(i), i))
            else:
                combos.append(i)
        return ("<h3>Singletons</h3> <ul> %s </ul><h3>The rest</h3> %s </p>"
                % (string.join(singletons," "),
                   present_nesting(self, combos, rel)))

    def __repr__(self):
        return ("<%s %s at 0x%s>"
                % (self.__class__.__name__, `self.id()`, hex(id(self))[2:]))


class WikiNesting(Acquisition.Implicit):
    """Given a wiki directory, generate nesting relationship from parents info.

    In a nesting, nodes are represented as:
       - Leaves: the string name of the page
       - Nodes with children: a list beginning with the parent node's name
       - Nodes with omitted children (for brevity): list with one string.
    """
    # XXX We could make this a persistent object and minimize recomputes.
    #     Put it in a standard place in the wiki's folder, or have the
    #     wikis in a folder share an instance, but use a single
    #     persistent one which need not recompute all the relationship
    #     maps every time - just needs to compare all pages parents
    #     settings with the last noticed parents settings, and adjust
    #     the children, roots, and parents maps just for those that
    #     changed.  On this first cut we just recompute it all...

    def __init__(self, container):
        self.container = container
        self.set_nesting()

    def set_nesting(self):
        """Preprocess for easy derivation of nesting structure.

        We set:
          - .parentmap: {'node1': ['parent1', ...], ...}
          - .childmap:  {'node1': ['child1', ...], ...}
          - .roots: {'root1': None, ...}"""

        pages = self.container.objectValues(spec='ZWiki Page')
        pagenames = []
        parentmap = {}                  # 'node': ['parent1', ...]
        childmap = {}                   # 'node': ['child1', ...]
        roots = {}
        for pg in pages:
            parents = pg.parents
            if type(parents) != ListType: # SKWM I think ken's zope remembers [] better
                parents = []
            pgnm = pg.id()
            pagenames.append(pgnm)

            # Parents is direct:
            parentmap[pgnm] = parents[:]

            # We have a root if node acknowledges no parents:
            if not parents:
                roots[pgnm] = None

            if not childmap.has_key(pgnm):
                childmap[pgnm] = []

            # Register page as child for all its parents:
            for p in parents:
                # We can't just append, in case we're a persistent object
                # (which recognizes the need to update by an attr assignment).
                if childmap.has_key(p): pchildren = childmap[p]
                else: pchildren = []
                if pgnm not in pchildren: pchildren.append(pgnm)
                childmap[p] = pchildren

        for k in parentmap.keys():
            parentmap[k].sort()
        for k in childmap.keys():
            childmap[k].sort()
        self.parentmap = parentmap
        self.childmap = childmap
        self.roots = roots

    def get_map(self):
        """Return nesting of entire wiki."""
        did = {}
        got = []
        roots = self.roots.keys()
        roots.sort()
        return self.get_offspring(roots)

    def get_offspring(self, pages, did=None):
        """Return nesting showing all offspring of a list of wiki pagenames.

        did is used for recursion, to prune already elaborated pages."""

        if did is None: did = {}
        got = []
        for p in pages:
            been_there = did.has_key(p)
            did[p] = None
            if self.childmap.has_key(p):
                children = self.childmap[p]
                if children:
                    subgot = [p]
                    if not been_there:
                        subgot.extend(self.get_offspring(children, did=did))
                    got.append(subgot)
                else:
                    got.append(p)
            else:
                got.append(p)
        return got

    def get_up_and_back(self, pagename):
        """Return nesting showing page containment and immediate offspring."""

        # Go up, identifying all and topmost forbears:
        ancestors = {}                  # Ancestors of pagename
        tops = {}                       # Ancestors that have no parents
        todo = {pagename: None}
        parentmap = self.parentmap
        while todo:
            doing = todo
            todo = {}
            for i in doing.keys():
                if ancestors.has_key(i):
                    continue            # We already took care of this one.
                else:
                    ancestors[i] = None
                    if parentmap.has_key(i):
                        parents = parentmap[i]
                    if parents:
                        for p in parents:
                            todo[p] = None
                    else: tops[i] = None

        ancestors[pagename] = None      # Finesse inclusion of page's offspring

        # Ok, now go back down, showing offspring of all intervening ancestors:
        tops = tops.keys()
        tops.sort
        did = {}; got = []
        childmap = self.childmap
        for t in tops:
            got.append(descend_ancestors(t, ancestors, did, childmap))
        return got

# --- geoffg's managed_mode code
def present_attributes(this_item):
    if this_item.managed_mode:
      this_owner = '<i>' + this_item.page_owner + '</i>'
      this_status = this_item.page_status
      # colour this in
      if this_status == 'review':
        this_status = '<font color=#00ccff>review</font>'
      elif this_status == 'approved':
        this_status = '<font color=#33cc00>approved</font>'
      elif this_status == 'rejected':
        this_status = '<font color=#ff6633>rejected</font>'
      s = ' ' + this_owner + ' ' + this_status
    else:
      s = ''
    return s
# end geoffg's managed_mode code
    
# geoffg's managed_mode code - I've put 'self' into the function below in place of 'myid'
def present_nesting(self, nesting, rel, did=None, _got=None, indent=""):
    """Present an HTML outline of a nesting structure.

    _got is used internally and should not be passed in except when recursing.

    See WikiNesting docstring for nesting structure description."""

    myid = self.id()
    if did is None: did = []
    if _got is None:
        _got = ["<ul>"]
        recursing = 0
    else:
        recursing = 1

    for n in nesting:

        if type(n) == ListType:
            if n[0] == myid:
                # The entry is the current node - distinguish it.
                nodenm = "<strong>%s</strong>" % n[0]
            else:
                nodenm = n[0]
            # --- geoffg's managed_mode code
            this_item = self.aq_parent[n[0]]
            # end geoffg's managed_mode code + use of function in line below
            _got.append('%s <li><a href="%s/%s">%s</a> %s'
                        % (indent, rel, quote(n[0]), nodenm, present_attributes(this_item)))
            if len(n) > 1:
                _got.append("<ul>")
                for i in n[1:]:
                    if type(i) == ListType:
                        _got = present_nesting(self, [i], rel, did=did,
                                               _got=_got, indent=indent+'  ')
                    else:
                        if i == myid:
                            # Distinguish that entry is for current node.
                            inm = "<strong>" + i + "</strong>"
                        else: inm = i
                        # --- geoffg's managed_mode code
                        this_item = self.aq_parent[i]
                        # end geoffg's managed_mode code + use of variable in line below
                        _got.append('%s <li><a href="%s/%s">%s</a> %s'
                                   % (indent, rel, quote(i), inm, present_attributes(this_item)))
                _got.append("</ul>")
            else:
                # Parents whose children were omitted - indicate:
                _got[-1] = _got[-1] + " ..."
        else:
            if n == myid:
                # The entry is for the current node - distinguish.
                inm = "<strong>" + n + "</strong>"
            else: inm = n
            # --- geoffg's managed_mode code
            this_item = self.aq_parent[n]
            # end geoffg's managed_mode code + use of variable in line below
            _got.append('%s <li><a href="%s/%s">%s</a> %s'
                       % (indent, rel, quote(n), inm, present_attributes(this_item)))

    if recursing:
        return _got
    else:
        _got.append("</ul>")
        return join(_got, "\n")

def descend_ancestors(page, ancestors, did, children):
    """Create nesting of ancestors leading to page.

    page is the name of the subject page.
    ancestors is a mapping whose keys are pages that are ancestors of page
    children is a mapping whose keys are pages children, and the values
       are the children's parents.

    Do not repeat ones we already did."""
    got = []
    for c in ((children.has_key(page) and children[page]) or []):
        if not ancestors.has_key(c):
            # We don't descend offspring that are not ancestors.
            got.append(c)
        elif ((children.has_key(c) and children[c]) or []):
            if did.has_key(c):
                # We only show offspring of ancestors once...
                got.append([c])
            else:
                # ... and this is the first time.
                did[c] = None
                got.append(descend_ancestors(c, ancestors, did, children))
        else:
            got.append(c)
    got.sort()                  # Terminals will come before composites.
    got.insert(0, page)
    return got
# ^^^ ken's parenting code
        

def manage_addZWikiPage(self, id, title='', file='', REQUEST=None,
                        submit=None):
    """Add a ZWiki Page object with the contents of file. If
    'file' is empty, default text is used.
    """
    if type(file) is not StringType: file=file.read()
    if not file:
        if hasattr(self, 'standard_wiki_page'):
            file = absattr(self.standard_wiki_page)
        else:
            file = default_wiki_page
    ob=ZWikiPage(file, __name__=id)
    ob.title=title
    id=self._setObject(id, ob)

    # need to add it to the catalog here too
    # will management changes, deletion update the catalog ?
    # do I need to define new properties screens with a catalog tab ?
    # ??

    if REQUEST is not None:
        try: u=self.DestinationURL()
        except: u=REQUEST['URL1']
        if submit==" Add and Edit ": u="%s/%s" % (u,quote(id))
        REQUEST.RESPONSE.redirect(u+'/manage_main')
    return ''

manage_addZWikiPageForm=HTMLFile('zwikipageAdd', globals())


# define these as dtml methods without the default_ prefix to override
# NB the methods will need <dtml-with this> to access the zwikipage vars

default_wiki_page=\
'''This is a new ZWiki page.
'''

default_wiki_header=HTML(source_string=\
'''<HTML>
<HEAD>
<TITLE>zwiki: <dtml-var id></TITLE>
</HEAD>
<BODY BGCOLOR="#FFFFFF">
<table width="100%" border="0" cellspacing="0" cellpadding="3"  bgcolor="eeeeee"><tr>
<td valign="middle">
<h1><dtml-var id></h1>
</td>
</tr>
</table>
<p>
<!-- end of standard_wiki_header -->

''')

default_wiki_footer=HTML(source_string=\
'''
<!-- start of standard_wiki_footer -->
<p>
<FORM METHOD="POST" ACTION="<dtml-var id url_quote>/edit" ENCTYPE="multipart/form-data">
<table width="100%" border="0" cellspacing="0" cellpadding="3" bgcolor="eeeeee">
<tr>
<td valign="center" align="center" colspan=2>
<TEXTAREA NAME="text" ROWS=4 COLS=70><dtml-var src></TEXTAREA>
</td>
</tr>
<tr>
<td valign="middle" align="left">
<small><em>
Last edited <dtml-var bobobase_modification_time>
</em></small>
</td>
<td valign="middle" align="right">
<INPUT TYPE="submit" NAME="Change" VALUE=" Change <dtml-var id> ">
<dtml-unless "AUTHENTICATED_USER.has_permission('Change ZWiki Pages',this())">
<br><em>
(changing this page<br>
will require a login)
</em>
</dtml-unless>
</td>
</tr>
</table>
</FORM>
</BODY>
</HTML>
''')

